from amaranth.lib import wiring, data
from amaranth.sim import Simulator
import random
from random import randrange
import pytest

from groovylight.fetcher import AddressConverter, AddressGenerator
from groovylight.geom import DisplayString, Coord, DisplayDimensions, DisplayRotation

ds_testdata = [
    (DisplayRotation.R0, (0, 0), {"x": 3, "y": 0}),
    (DisplayRotation.R0, (40, 2), {"x": 43, "y": 2}),
    (DisplayRotation.R90, (40, 0), {"x": 66, "y": 40}),
    (DisplayRotation.R90, (120, 2), {"x": 64, "y": 120}),
]


@pytest.mark.parametrize("rot, inp, expected", ds_testdata)
def test_converter(rot, inp, expected):
    ds = DisplayString(Coord(3, 0), DisplayDimensions(128, 64), rot)
    dut = AddressConverter(ds)
    sim = Simulator(dut)

    async def testbench(ctx):
        await ctx.delay(1e-6)
        ctx.set(dut.input_x, inp[0])
        ctx.set(dut.addr, inp[1])
        await ctx.delay(1e-6)
        assert ctx.get(dut.output)[0]["x"] == expected["x"]
        assert ctx.get(dut.output)[0]["y"] == expected["y"]

    sim.add_testbench(testbench)

    with sim.write_vcd("output.vcd"):
        sim.run()


# Helper functions for stream management
async def stream_get(ctx, stream):
    ctx.set(stream.ready, 1)
    (payload,) = await ctx.tick().sample(stream.payload).until(stream.valid)
    ctx.set(stream.ready, 0)
    return payload


async def stream_put(ctx, stream, payload):
    ctx.set(stream.valid, 1)
    ctx.set(stream.payload, payload)
    await ctx.tick().until(stream.ready)
    ctx.set(stream.valid, 0)


generator_tests = [
        (0, DisplayRotation.R0),
        (0, DisplayRotation.R90),
        (4, DisplayRotation.R90),
        
        ]

@pytest.mark.parametrize("addr, rot", generator_tests)
def test_generator(addr, rot):
    ds = DisplayString(Coord(3, 0), DisplayDimensions(128, 64), rot)
    dut = AddressGenerator(ds)
    sim = Simulator(dut)
    sim.add_clock(1e-6)

    async def runner(ctx):
        # TODO: set inputs
        ctx.set(dut.addr, addr)
        await ctx.tick()
        ctx.set(dut.start, 1)
        await ctx.tick()
        ctx.set(dut.start, 0)
        await ctx.tick().until(dut.done == 1)

    expected = [
        [ds.translate_coord(x, addr, 0), ds.translate_coord(x, addr, 1)] for x in range(128)
    ]

    expected.reverse()

    async def stream_checker(ctx):
        while ctx.get(dut.done) == 0:
            payload = await stream_get(ctx, dut.coordstream)
            assert expected.pop() == payload

    sim.add_testbench(runner)
    sim.add_testbench(stream_checker)
    with sim.write_vcd("generator.vcd"):
        sim.run()
